#
# The Authors:
# Andre Teves - andreteves@gmail.com
# Danillo Couto - danillo_couto@yahoo.com
#
# Support:
# FAPESP (www.fapesp.br)
# Escola Politecnica - University of Sao Paulo (www.poli.usp.br)
# Decision Making Lab - directed by Fabio Cozman (http://www.pmr.poli.usp.br/ltd/)
#

"""Propositional Probabilistic Logic Package

PPL is a convenient editor for PSAT problems that simplifies
the construction of knowledge bases, checks its coherence
and performs extensions on it.

Full Compatibility with Python 2.4

For more information visit:
http://www.pmr.poli.usp.br/ltd/Software/ppl/
"""

#libs
import re
from to_CNF import logic
import os 
import subprocess
 

#global variables
assoc_dic = {}
clauses_dic = {}
clauses_num_dic = {}
prob_dic = {}
resultado = {}
clauses_num_list = []
i = 1
j = 1
numcnf = 0
s1 = ''
sent_inferir = ''
filename = "ppl_file.txt"

pplpath = os.environ["PPLPATH"]
geracolpath = os.environ["PPLPATH"] + "\geracol"


def toCNF(s):
    """This functions uses the module to_CNF, which was developed by
    AIMA Online Code Repository. It was based in the pseudocode presented in
    page 215 of the book Artificial Intelligence: A Modern Approach, written
    by Stuart J. Russell and Peter Norvig.
    
    This module can be found at http://aima.cs.berkeley.edu/code.html
    
    It Converts a propositional logical sentence s to conjunctive normal form.
    >>> to_cnf("~(B|C)")
    (~B & ~C)
    >>> to_cnf("B <=> (P1|P2)")
    ((~P1 | B) & (~P2 | B) & (P1 | P2 | ~B))
    >>> to_cnf("a | (b & c) | d")
    ((b | a | d) & (c | a | d))
    """

    a = logic.to_cnf(s)
    return str(a)


def p(s, pmin = '', pmax = ''):
   """This is the main fuction of PPL. It allows the user to create the
   knowledge base by the insertion of sentences and the association of
   either probabilities or probability intervals to them.
    
   There is a specific sintax that must be followed in order to insert
   a sentence. In general, a sentence is a set of operators and args.
   The possible operators are:

   Null-ary (no args) op:
      A symbol, representing a variable or constant (e.g. 'a')
   Unary (1 arg) op:
      '~', '-', representing NOT, negation (e.g. '~a')
   Binary (2 arg) op:
      '==>' or '>>', representing forward implication
      '<==' or '<<', representing backward implication
      '<=>' or '%', representing logical equality
      '=/=' or '^', representing logical disequality (XOR)
   N-ary (0 or more args) op:
      '&', '|', representing conjunction and disjunction

   For more information about the sintax see PPL manual.
   """

   global s1

   arq = ''
   arq_final = ''

   splitInClauses(s)
   toNumber()

   if pmax == '':
       pmax = pmin

   if pmin > pmax:
        temp = pmax
        pmax = pmin
        pmin = temp

   for x in s1:
       arq = arq +  str(clauses_dic[' '+x.lstrip('( ').rstrip(') ')+' ']) + '  '

   if pmin != '':
       arq_final = str(pmin) + ' ' + str(pmax) + '   ' + arq
       prob_dic[arq.rstrip()] = [str(pmin), str(pmax)]


def splitInClauses(s):
   """Split a given sentence s in its clauses, which is a disjunction
   of literals.
   """
   
   global s1
   global j
   
   sent = toCNF(s)
   s1 = sent.split('&')
   for x in s1:
       if clauses_dic.has_key(' '+x.lstrip('( ').rstrip(') ')+' ') == 0:
           clauses_dic[' '+x.lstrip('( ').rstrip(') ')+' '] = j
           j = j + 1


def mapToNumber():

   global i
   global s1
   
   var = []

   for x in s1:
       var = var + [(x.split('|'))]

   for y in var:
       for x in y:
           if assoc_dic.has_key(' '+x.lstrip('~( ').rstrip(') ')+' ') == 0:
               assoc_dic[' '+x.lstrip('~( ').rstrip(') ')+' '] = i
               i = i + 1


def toNumber():

   mapToNumber()

   for x in clauses_dic:
       xorig = x

       x = x.replace('~', '- ')
       
       for y in assoc_dic:
           if x.find(y) != -1:
               x = x.replace( x[(x.find(y)):(x.find(y))+len(y)], str(assoc_dic[y]) )

       for z in x:
           if z.isalnum() == 0 and z != '-':
               x = x.replace(z,' ')

       x = x.replace('- ', '-')
       x = x.replace('  ',' ')

       clauses_num_dic[x] = clauses_dic[xorig]

       
def toOrderedList(dic):
    """Transform a given dictionary in an ordered list. 
    """

    lista = []
    
    item = dic.items()

    for (key, val) in item:
        lista = lista + [(val, key)]
    
    lista.sort()
    
    return lista


def toFile():
    """Create a txt file with all the information inserted by the user. This
    file can be found in directory geracol.
    This file is used by Geracol to check consistency and perform extensions.
    """

    global clauses_num_list
    global pplpath
    global geracolpath
    global sent_inferir
    global numcnf
    global filename

    f = open(geracolpath.lstrip() + "\\" + filename, 'w')

    numvar = i - 1
    numcla = j - 1
    numcnf = numcnf + len(prob_dic)

    f.write(str(numvar) + ' ' + str(numcla) + ' ' + str(numcnf) + '\n\n')

    clauses_num_list = toOrderedList(clauses_num_dic)

    for (key, val) in clauses_num_list:
        f.write(val.lstrip() + ' 0\n')

    f.write('\n')

    for x in prob_dic:
        for y in prob_dic[x]:
            f.write(y + ' ')
        f.write(' ' + x + ' ' + '0\n')

    f.write(sent_inferir + '\n')

    f.close()

   
def launchWithoutConsole(command, args):
    
    startupinfo = subprocess.STARTUPINFO()
    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    return subprocess.Popen([command] + args, startupinfo=startupinfo).wait()


def callGeracol():
    """To check consistency and perform extensions, PPL executes
    calls to the consistency checker called Geracol.

    Geracol is an implementaion of the column generation method
    (Hansen, P., Jaumard, B., and de Aragao, M. (1991). Column
    generation methods for probabilistic logic. ORSA Journal
    on Computing 3:135-148).

    It was developed by Paulo Andrade in C language.
    """
    
    global geracolpath
    global pplpath
    global filename
    
    os.chdir(geracolpath)
    launchWithoutConsole("geracol.exe", [filename])


def loadFile():
    
    global pplpath
    global resultado

    os.chdir(geracolpath)
    f = open("saida.txt")

    resultado["HINI"] = findWords(f, "HINI")
    resultado["RSL"] = findWords(f, "RSL")
    resultado["MIN"] = findWords(f, "MIN")
    resultado["MAX"] = findWords(f, "MAX")
    resultado["HFIM"] = findWords(f, "HFIM")
    
    f.close()


def findWords(f, word):
    """Return the lastline of a file f that have a certain word in it."""
    
    f.seek(0,0)
    for line in f:
        if re.search(r"\b" + word + r"\b", line):
            lastline = line
    return lastline

             
def checkCoherence():

    toFile()
    callGeracol()
    loadFile()

    if resultado["RSL"].split()[1] == "S":
        return "coherent!"
    if resultado["RSL"].split()[1] == "N":
        return "incoherent!"
    if resultado["RSL"].split()[1] == "F":
        return "ERROR - Geralcol failed to process given database"

    
def extension(sent):

   global sent_inferir
   global numcnf

   arq = ''
   arq_final = ''

   numcnf = numcnf + 1
   splitInClauses(sent)
   toNumber()

   for x in s1:
       arq = arq +  str(clauses_dic[' '+x.lstrip('( ').rstrip(') ')+' ']) + '  '

   sent_inferir = str(0.0) + ' ' + str(0.0) + '   ' + arq + ' 0\n'

   status = checkCoherence()
   if status == "coherent!":
       print "Lower: " + resultado["MIN"].split()[1]
       print "Upper: " + resultado["MAX"].split()[1]
   else:
       print status
