/* Arquivo: geracol.c
// Versao 1.1
//
// Autor: Paulo Sergio de Souza Andrade                n.usp: 4963109
// Fevereiro de 2007
//
// geracol.c 
// Aplicacao para inferencia de um valor min/max de probabilidade para
// uma sentenca adicional, dado como funcao de um banco de dados
// com um conjunto de sentencas na forma normal conjuntiva e seus
// valores de probabilidade dados na forma de intervalos.
// Implementa o metodo de geracao de colunas e utiliza o programa
// de otimizacao ILOG CPLEX para resolucao dos problemas envolvidos
// com o metodo.
// Recomenda-se a leitura de formato.pdf para esclarecimento do
// argumento que deve ser passado para o programa, e de Leia-me,
// para instrucoes de compilacao e notas sobre saidas.
/////////////////////////////////////////////////////////////////// */

/* Bibliotecas Utilizadas
/////////////////////////////////////////////////////////////////// */
// a primeira inclui stdio.h juntamente com cplex.h
#include </usr/ilog/cplex100/include/ilcplex/cplex.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

/* Constantes
/////////////////////////////////////////////////////////////////// */
# define TRUE 1       // para utilizacao em lacos
# define NOMEMAX 1000 // comprimento admitido para nomear variaveis
# define MAXLIN 10    // quebra de linha
# define EPS 1.0e-6   // erro admitido
# define PAUSA 1.0e+5 // para aguardar liberacao da licenca

/* Prototipos de Funcoes
/////////////////////////////////////////////////////////////////// */
int preparafase1 (CPXCENVptr env, CPXLPptr lp, int numcnf,
                  int *indcoef, int numcoef, double *valorcoef);
int otimizalp (CPXCENVptr env, CPXLPptr lp, double *objval_p, 
               double *p, double *pi);
int otimizamip (CPXCENVptr env, CPXLPptr mip, int numvar, int numcla, 
                int numcnf, double *x, int *indices, double *pi,
                double *objval_p, int *mundo, double *colentr);
int atualizalp (CPXCENVptr env, CPXLPptr lp, int numcnf, int numvar, 
                double coefobj, double *colentr, int *indlin,
                int *mundo, char **nomecol);
int preenchelp (CPXENVptr env, CPXLPptr lp, int numcnf, double *prob,
                double *probsup);
int preenchemip (CPXENVptr env, CPXLPptr mip, int numvar, int numcla, 
                 int numcnf, int **CLA, int *tamcla, int **CNF,
                 int *tamcnf);
int ledados (char *nomearq, int *numcla_p, int *numvar_p,
             int *numcnf_p, int ***CLA_p, int **tamcla_p,
             int ***CNF_p, int **tamcnf_p,double **prob_p,
             double **probsup_p);
FILE *limpa (FILE *in, FILE *arq);
void desaloca (char **ptr);

/* Programa Propriamente Dito
/////////////////////////////////////////////////////////////////// */
int main (int numargs, char **v) {  
  int status = 0;
  int i, j;

  /* Dados associados ao arquivo de entrada */
  int numcla;          // numero de clausulas,
  int numvar;          // de variaveis e
  int numcnf;          // de expressoes com probabilidade
  int **CLA = NULL;    // CLA[i][] contera a i-esima clausula
  int *tamcla = NULL;  // tamcla[i] sera o numero de variaveis
                       // que compoem a i-esima clausula
  int **CNF = NULL;    // CNF[i][] contera a i-esima CNF
  int *tamcnf = NULL;  // tamcnf[i] sera o numero de clausulas
                       // que compoem a i-esima CNF
  double *prob = NULL; // prob[i] sera a probabilidade da i-esima CNF
  double *probsup = NULL; // prob[i] sera a prob. sup. da i-esima CNF

  /* Cria ponteiros para ambiente e problemas do cplex */
  CPXENVptr env = NULL; // ambiente
  CPXLPptr lp = NULL;   // problema de programacao continua
  CPXLPptr mip = NULL;  // problema de programacao inteira

  /* Vetores e variaveis relacionadas ao problema PMR (= lp) */
  double objval;      // valor otimo do lp
  double *p = NULL;   // vetor solucao de valores do primal
  double *pi = NULL;  // vetor de valores do dual
  int itlp = 0;       // numero de iteracoes para resolver lp
  // para coluna de entrada
  int *indlin = NULL;     // indices das linhas
  double coefobj = 0;     // coeficiente da objetiva
  double *colentr = NULL; // coluna de entrada
  char **nomecol = NULL;  // nome da coluna de entrada
  int *indcoef = NULL;    // atuam na fase0
  double *valorcoef = NULL;
  
  /* Vetores e variaveis relacionadas ao SP (= mip) */
  double objvalmip;    // valor otimo
  double *x = NULL;    // vetor solucao
  int itmip = 0;       // numero de iteracoes para resolver mip
  int *mundo = NULL;   // atribuicao dada as variaveis {0, 1}
  int *indices = NULL; // indices para substituir objetiva
  double atua = 1.0;   // redefine coefobj para sent. adicional
  int indatua;         // indice para alterar coefobj acima 

  int itgc = 0; // numero de iteracoes do metodo de geracao de colunas
  int itex = 0; // numero de vezes que SP e resolvido exatamente

  /* Variaveis para controle do tempo aproximado de execucao */
  double fim;
  double inicio;
  double tempototallp = 0;
  double tempototalmip = 0;
  time_t t;
  char *horario;

  /* Para arquivo de escrita */
  FILE *saida = NULL;
  char *arquivo = "saida.txt";

  /* Informacoes iniciais para arquivo de saida */
  saida = fopen (arquivo, "a");
  if (saida == NULL) {
    printf ("** ERRO ** Arquivo de escrita nao encontrado.\n");
    status = 1;
    goto FINAL;
  }
  t = time(NULL);
  horario = asctime (localtime (&t));
  fprintf (saida,"\nINICIO \nHINI %sARQ %s", horario, v[1]);

  /* Verifica argumentos da linha de comando */
  if (numargs != 2) {
    printf ("** ERRO ** Lista de argumentos errada.\n");
    goto FINAL;
  }
  
  /* Le arquivo de entrada */
  status = ledados (v[1], &numcla, &numvar, &numcnf, &CLA,
                    &tamcla, &CNF, &tamcnf, &prob, &probsup);
  if (status) goto FINAL;
  // nao preciso da linha abaixo pois tenho uma sent. adicional
  // que foi contabilizada dado que o problema apresenta uma restricao
  // adicional 1*x = 1
  //numcnf++;
  
  /* Alocacoes */
  p = (double *) malloc (numcnf*sizeof (double));
  pi = (double *) malloc (numcnf*sizeof (double)); 
  indlin = (int *) malloc (numcnf*sizeof (int));
  colentr = (double *) malloc ((numcnf+1)*sizeof (double));
  nomecol = (char **) malloc (sizeof (char *));
  x = (double *) malloc ((numvar+numcla+numcnf)*sizeof (double));
  mundo = (int *) malloc (numvar*sizeof (int));
  indices = (int *) malloc ((numcnf-1)*sizeof (int));
  // Suponho que a fase0 precisara de no maximo numcnf*MAXLIN
  // iteracoes, logo aloco os seguintes vetores
  indcoef = (int *) malloc ((numcnf*MAXLIN)*sizeof (int));
  valorcoef = (double *) malloc ((numcnf*MAXLIN)*sizeof (double));
  if (p == NULL || pi == NULL || indlin == NULL || colentr == NULL ||
      nomecol == NULL || x == NULL || mundo == NULL ||
      indices == NULL || indcoef == NULL || valorcoef == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  nomecol[0] = (char *) malloc (NOMEMAX*sizeof (char));
  if (nomecol[0] == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }

  /* Inicializa ambiente cplex */
  while (TRUE) {
    env = CPXopenCPLEX (&status);
    if (status == 0) break;
    for (i = 0; i < PAUSA; i++);
  }
  
  /* Desliga saida do CPLEX para tela */
  status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_OFF);
  if (status) {
    printf ("Falha ao desligar saida para tela, erro %d.\n", status);
    goto FINAL;
  }

  /* Cria os objetos problema */
  lp = CPXcreateprob (env, &status, "PMR");
  if (lp == NULL) {
    printf ("Falha ao criar LP.\n");
    goto FINAL;
  }

  mip = CPXcreateprob (env, &status, "Subprob");
  if (mip == NULL) {
    printf ("Falhou para criar MIP.\n");
    goto FINAL;
  }

  /* Preenche dados iniciais dos problemas */
  status = preenchelp (env, lp, numcnf, prob, probsup) ;
  if (status) {
    printf ("Falhou para preencher LP.\n");
    goto FINAL;
  }

  status = preenchemip (env, mip, numvar, numcla, numcnf, CLA,
                        tamcla, CNF, tamcnf);
  if (status) {
    printf ("Falhou para preencher MIP.\n");
    goto FINAL;
  }
  
  // indices dos pesos para mip alterar objetiva
  for (i = 0; i < numcnf-1; i++)
    indices[i] = numvar+numcla+i;
  // indices das linhas para adicionar coluna no lp
  for (i = 0; i < numcnf; i++)
    indlin[i] = i;
  
  /* Altera parametro para resolver mip aproximadamente */
  status = CPXreadcopyparam (env, "altparam.prm");
  if (status) {
    printf ("Falha ao mudar parametros, erro %d.\n", status);
    goto FINAL;
  }

  /* Realiza a fase0 quando i == 0
  // Caso viavel:
  // minimiza para i == 1, fornecendo limite inferior de prob.
  // maximiza para i == 2, fornecendo limite superior de prob.
  ///////////////////////////////////////////////////////////// */
  for (i = 0; i < 3; i++) {
    /* Metodo numerico de geracao de colunas */
    while (TRUE) {
      itgc++;

      /* Otimiza LP e obtem solucao */
      inicio = (double) clock ();
      status = otimizalp (env, lp, &objval, p, pi);
      fim = (double) clock ();
      tempototallp += fim - inicio;
      if (status) {
        printf ("Falhou para otimizar LP.\n");
        goto FINAL;
      }
      // obtem numero de iteracoes para resolver LP
      itlp += CPXgetitcnt (env, lp);
  
      /* Verifica se atingiu 0, implicando problema viavel */
      if (i == 0 && objval < EPS) {
          printf ("Problema e viavel.\n");
          fprintf (saida,"\nRSL S");
          break;
      }
           
      /* Otimiza MIP e obtem solucao */
      inicio = (double) clock ();
      status = otimizamip (env, mip, numvar, numcla, numcnf,  x,
                           indices, pi, &objvalmip, mundo, colentr);
      fim = (double) clock ();
      tempototalmip += fim - inicio;
      if (status) {
        printf ("Falhou para otimizar MIP.\n");
        goto FINAL;
      }
      // obtem numero de iteracoes para resolver MIP
      itmip += CPXgetmipitcnt (env, mip);
      
      /* Comprova valor otimo */
      if ((objvalmip > -EPS && (i == 0 || i == 1)) ||
          (objvalmip < EPS && i == 2)) {
        itex++;
	
        /* Altera parametro para definicao padrao */
        status = CPXreadcopyparam (env, "norparam.prm");
        if (status) {
          printf ("Falha ao redefinir parametros, erro %d.\n", status);
          goto FINAL;
        }

        /* Realtera multiplicadores */
        for (j = 0; j < numcnf; j++)
          pi[j] = (-1)*pi[j];

        /* Otimiza MIP e obtem solucao */
        inicio = (double) clock ();
        status = otimizamip (env, mip, numvar, numcla, numcnf, x,
                             indices, pi, &objvalmip, mundo, colentr);
        fim = (double) clock ();
        tempototalmip += fim - inicio;
        if (status) {
          printf ("Falhou para otimizar MIP.\n");
          goto FINAL;
        }
        // obtem numero de iteracoes para resolver MIP
        itmip += CPXgetmipitcnt (env, mip);
	    
	/* Interpreta resultado */
        // fase0 verificando inviabilidade
        if (i == 0 && objvalmip > -EPS) {
          printf ("Problema e inviavel.\n");
          fprintf (saida,"\nRSL N");
          goto FINAL;
        }
        // finaliza minimizacao
        else if (i == 1 && objvalmip > -EPS) {
          printf ("limite inferior = %g\n", objval); 
          fprintf (saida,"\nMIN %e", objval);
          break;
        }
        // finaliza maximizacao
        else if (i == 2 && objvalmip < EPS){
          printf ("limite superior = %g\n", objval); 
          fprintf (saida,"\nMAX %e", objval);
          break;  
        } 
        
        /* Realtera parametro para resolver mip aproximadamente */
        status = CPXreadcopyparam(env, "altparam.prm");
        if (status) {
          printf ("Falha ao mudar parametros, erro %d.\n",status);
          goto FINAL;
        }
      }
  
      /* Coleta indice do coeficiente da objetiva */
      // na fase0, guardo para que na fase1 voltem
      if (i == 0) {
       indcoef[itgc-1] = CPXgetnumcols (env, lp);
       valorcoef[itgc-1] = colentr[numcnf];
      } 
      // coefobj so foi zero na fase0
      else coefobj = colentr[numcnf]; 
      
      /* Insere coluna no LP */
      status = atualizalp (env, lp, numcnf, numvar, coefobj,
                           colentr, indlin, mundo, nomecol);
      if (status) {
        printf ("Falhou para atualizar lp.\n");
        goto FINAL;
      }
    }

    /* Termino da fase0 com sucesso => problema viavel, aplico a fase
    // seguinte, minimizacao seguida de maximixacao */
    if (i == 0) {
      status = preparafase1 (env, lp, numcnf, indcoef, itgc-1,
                             valorcoef); 
      if (status) {
        printf ("Erro ao preparar problema para fase1.\n");
        goto FINAL;
      }

      /* Altera coeficientes da objetiva para mip considerar
      // sentenca adicional */
      indatua = numvar+numcla+numcnf-1;
      status = CPXchgobj (env, mip, 1, &indatua, &atua);
      if (status) {
	printf ("Falhou para mudar objetiva do MIP.\n");
	goto FINAL;
      }
    }
    else {
      CPXchgobjsen (env, lp, CPX_MAX);
      CPXchgobjsen (env, mip, CPX_MAX);     
    }
  }
  
  FINAL:

  ////////////////////////////////////////////////////////////
  /* Curiosidade
  status = status = CPXwriteprob (env, lp, "proPMR.lp", NULL);
  if (status)
    printf ("Falhou para obter arquivo com problema LP.\n");
  status = CPXmbasewrite (env, lp, "basePMR.bas");
  if (status)
    printf ("Falhou para obter arquivo com base.\n");
  status = CPXsolwrite (env, lp, "solPMR.sol");
  if (status)
    printf ("Falhou para obter arquivo com solucao.\n");  
  //////////////////////////////////////////////////////////
  
  /* Interrupcao da fase0 por algum motivo, aconselhavel
  // verificar informacao enviada para tela */
  if (status)
    fprintf (saida,"\nRSL F");

  /* Envia informacoes finais ao arquivo de saida */
  fprintf (saida, "\nTMPC %e", tempototallp/CLOCKS_PER_SEC);
  fprintf (saida, "\nTMPI %e", tempototalmip/CLOCKS_PER_SEC);
  fprintf (saida, "\nITEC %d", itlp);
  fprintf (saida, "\nITEI %d", itmip);
  fprintf (saida, "\nITGC %d", itgc);
  fprintf (saida, "\nITEX %d", itex);
  /* Desaloca problemas */
  if (lp != NULL) {
    status = CPXfreeprob (env, &lp);
    if (status) {
      printf ("CPXfreeprob falhou, codigo de erro %d.\n", status);
    }
  }

  if (mip != NULL) {
    status = CPXfreeprob (env, &mip);
    if (status) {
      printf ("CPXfreeprob falhou, codigo de erro %d.\n", status);
    }
  }

  /* Desaloca ambiente */
  if (env != NULL) {
    status = CPXcloseCPLEX (&env);
    if (status) {
      char  errmsg[1024];
      printf ("Falha ao fechar o ambiente CPLEX.\n");
      CPXgeterrorstring (env, status, errmsg);
      printf ("%s", errmsg);
    }
  }

  /* Desaloca vetores e matrizes de dados */
  desaloca ((char **) &p);
  desaloca ((char **) &pi);
  desaloca ((char **) &indlin);
  desaloca ((char **) &colentr);
  desaloca ((char **) &nomecol[0]);
  desaloca ((char **) &nomecol);
  desaloca ((char **) &x);
  desaloca ((char **) &mundo);
  desaloca ((char **) &indices);
  desaloca ((char **) &indcoef);
  desaloca ((char **) &valorcoef);
  if (CLA != NULL) {
    for (i = 0; i < numcla; i++)
      desaloca ((char **) &CLA[i]);
  }
  desaloca ((char **) &CLA);
  if (CNF != NULL) {
    for (i = 0; i < numcnf; i++)
      desaloca ((char **) &CNF[i]);
  }
  desaloca ((char **) &CNF);
  desaloca ((char **) &tamcla);
  desaloca ((char **) &tamcnf);
  desaloca ((char **) &prob);
  desaloca ((char **) &probsup);

  /* Informacoes finais para arquivo de saida */
  t = time(NULL);
  horario = asctime(localtime(&t));
  fprintf (saida,"\nHFIM %sFIM", horario);

  fclose (saida);

  return EXIT_SUCCESS;
}

/* Altera funcao objetivo com os coeficientes de colunas geradas, que
// haviam ficado de fora durante a fase0, sendo que o valor dos mesmos
// esta em valorcoef[1...numcoef] e seus indices em indcoef, em
// seguida elimina de lp, variavel nao basica, diferente de variavel
// de folga.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int preparafase1 (CPXCENVptr env, CPXLPptr lp, int numcnf,
                  int *indcoef, int numcoef, double *valorcoef) {
  int status = 0;
  int numcol;  // numero de colunas no problema
  int tiposol;  // tipo de solucao, basica ou nao basica...
  int *cstat = NULL;  // status da coluna, basica ou nao basica
  int *delstat = NULL;  // marca colunas que serao eliminadas
  char **nomecol = NULL;  // nome da variavel
  char *tamnome = NULL;  // comprimento associado a nome de variavel
  int espaconome, folga; // associados ao espaco ocupado por nomes
  char infsup = 'B'; // define tipo fixo
  double fixo = 0; // valor que assumira variavel artificial
  int i = 0;
  
  status = CPXsolninfo (env, lp, NULL, &tiposol, NULL, NULL);
  if (status) {
    printf ("Falha ao tentar obter informacao sobre solucao.\n");
    status = 1;
    goto FINAL;
  }
  numcol = CPXgetnumcols (env, lp);
  if (tiposol == CPX_BASIC_SOLN) {
    cstat = (int *) malloc (numcol*sizeof (int));
    delstat = (int *) malloc (numcol*sizeof (int));
    if (cstat == NULL || delstat == NULL) {
      printf ("** ERRO ** Malloc devolveu NULL!\n");
      status = 1;
      goto FINAL;
    }
    status = CPXgetbase (env, lp, cstat, NULL);
    if ( status ) {
      fprintf (stderr, "Falha ao obter base; erro %d.\n", status);
      goto FINAL;
    }
  }
  else {
    printf ("Nao ha base\n");
    status = 1;
    goto FINAL;
  }

  // reconstroi funcao objetivo para fase1
  status = CPXchgobj (env, lp, numcoef, indcoef, valorcoef);
  if (status) {
    printf ("Falha ao mudar coeficientes da objetiva.\n");
    status = 1;
    goto FINAL;
  }

  // realiza um chato procedimento para obter nomes
  status = CPXgetcolname (env, lp, NULL, NULL, 0, &folga, 0,numcol-1);
  if ((status != CPXERR_NEGATIVE_SURPLUS) && (status != 0)) {
    printf ("Falha ao determinar espaco para nomes de colunas.\n");
    status = 1;
    goto FINAL;
  }
  espaconome = -folga;
  nomecol = (char **) malloc (numcol*sizeof(char *));
  tamnome = (char *) malloc (espaconome);
  if (nomecol == NULL || tamnome == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  status = CPXgetcolname (env, lp, nomecol, tamnome,
                          espaconome, &folga, 0, numcol-1);
  if (status) {
    printf ("Falha ao obter nomes de colunas.\n");
    status = 1;
    goto FINAL;
  }

  // identifica variavel nao basica e diferente de var. de folga  
  for (i = 0; i < numcol; i++) {
    if ((cstat[i] != CPX_BASIC) && (nomecol[i][0] != 'z'))
      delstat[i] = 1;
    else delstat[i] = 0;
  }     
  // elimina colunas
  status = CPXdelsetcols (env, lp, delstat);
  if (status) {
    printf ("Falha ao eliminar colunas.\n");
    status = 1;
    goto FINAL;
  }
  // desaloca variaveis relacionadas a nomes
  if (nomecol != NULL) {
    for (i = 0; i < numcol; i++)
      desaloca ((char **) &nomecol[i]);
  }
  desaloca ((char **) &nomecol);

  // redefine limites e coeficiente da funcao objetivo sobre variavel
  // artificial caso a mesma esteja na base
  numcol = CPXgetnumcols (env, lp);
  // volto a questionar sobre os nomes
  status = CPXgetcolname (env, lp, NULL, NULL, 0, &folga, 0,numcol-1);
  if ((status != CPXERR_NEGATIVE_SURPLUS) && (status != 0)) {
    printf ("Falha ao determinar espaco para nomes de colunas.\n");
    status = 1;
    goto FINAL;
  }
  espaconome = -folga;
  nomecol = (char **) malloc (numcol*sizeof(char *));
  tamnome = (char *) malloc (espaconome);
  if (nomecol == NULL || tamnome == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  status = CPXgetcolname (env, lp, nomecol, tamnome, 
                          espaconome, &folga, 0, numcol-1);
  if (status) {
    printf ("Falha ao obter nomes de colunas.\n");
    status = 1;
    goto FINAL;
  }
  // para entao identificar variaveis artificiais na base
  for (i = 0; i < numcol; i++) {
    if (nomecol[i][0] == 'y') {
      status = CPXchgobj (env, lp, 1, &i, &fixo);
      if (status) {
	printf ("Falha ao mudar valor objetivo.\n");
	status = 1;
	goto FINAL;
      }
      status = CPXchgbds (env, lp, 1, &i, &infsup, &fixo);
      if (status) {
	printf ("Falha ao mudar limites.\n");
	status = 1;
	goto FINAL;
      }
    }
  }

  FINAL:
  desaloca ((char **) &cstat); 
  desaloca ((char **) &delstat);
  if (nomecol != NULL) {
    for (i = 0; i < numcol; i++)
      desaloca ((char **) &nomecol[i]);
  }
  desaloca ((char **) &nomecol);
  desaloca ((char **) &tamnome);
  return status;     
}

/* Otimiza lp, fornecendo o valor otimo da objetiva em objval_P, o
// vetor com os valores das variaveis que compoem a base p, e o vetor
// de valores do dual pi, do problema lp no ambiente env.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */ 
int otimizalp (CPXCENVptr env, CPXLPptr lp, double *objval_p,
               double *p, double *pi) {
  int status = 0;
  int solstat; // status da solucao

  /* Otimiza LP e obtem solucao */
  status = CPXlpopt (env, lp);
  if ( status ) {
    printf ("Falhou para otimizar LP.\n");
    goto FINAL;
  }
  /* OBS: Obter numero de colunas e alocar p caso
  // haja interesse em obte-lo */
  status = CPXsolution (env, lp, &solstat, objval_p,
                        NULL, pi, NULL, NULL);
  if ( status ) {
    printf ("Falhou para obter solucao.\n");
    goto FINAL;
  }
  // interpreta status da solucao
  if (solstat == CPX_STAT_OPTIMAL || solstat == CPX_STAT_NUM_BEST);
  else {
    printf ("Solucao otima nao esta avaliada.\n");
    status = 1;
    goto FINAL;
  }
  
  FINAL:
  return status;
}

/* Otimiza subproblema mip, porem antes altera coeficientes da funcao
// objetivo com os vetores pi[0..nuncnf-1] e indices[0..numcnf-2],
// e fornece o valor otimo da objetiva em objval_p, a solucao otima
// titulada no vetor x[0..(numvar+numcla+numcnf)-1] e a que define o
// mundo possivel no vetor  mundo[0..numvar-1] presentes no ambiente
// env.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */ 
int otimizamip (CPXCENVptr env, CPXLPptr mip, int numvar, int numcla,
                int numcnf, double *x, int *indices, double *pi,
                double *objval_p, int *mundo, double *colentr) {
  int status = 0;
  int solstat;    // status da solucao
  int numcol;     // numero de colunas no problema
  double cte = 0; // constante associada ao mip
  int i, aux;

  /* Prepara multiplicadores */
  for (i = 0; i < numcnf; i++)
    pi[i] = (-1)*pi[i];
  // constante
  cte = pi[numcnf-1];

  /* Altera coeficientes da funcao objetivo */
  status = CPXchgobj (env, mip, numcnf-1, indices, pi);
  if (status) {
    printf ("Falhou para mudar MIP.\n");
    goto FINAL;
  }

  /* Otimiza o problema e obtem solucao */
  status = CPXmipopt (env, mip);
  if (status) {
    printf ("Falhou para otimizar MIP.\n");
    goto FINAL;
  }
  solstat = CPXgetstat (env, mip);
  // interpreta status da solucao
  if (solstat == CPXMIP_OPTIMAL || solstat == CPXMIP_OPTIMAL_TOL ||
      solstat == CPXMIP_SOL_LIM);
  else {
    printf ("Solucao otima nao esta avaliada.\n");
    status = 1;
    goto FINAL;
  }
  status = CPXgetobjval (env, mip, objval_p);
  if (status) {
    printf ("MIP nao forneceu valor objetivo.\n");
    goto FINAL;
  }
  // soma constante que havia ficado de fora 
  *objval_p += cte;
     
  // obtem numero de colunas, interessante quando nao sabemos!
  numcol = CPXgetnumcols (env, mip);
  status = CPXgetx (env, mip, x, 0, numcol-1);
  if (status) {
    printf ("Falhou para obter solucao inteira.\n");
    goto FINAL;
  }
  /* Obtem mundo possivel resultante */
  for (i = 0; i < numvar; i++)
    mundo[i] = (int) (x[i] + EPS);

  /* Obtem coluna de entrada resultante */
  aux = numvar+numcla;
  for (i = 0; i < numcnf; i++)
    colentr[i] = (double) ((int) (x[i+aux] + EPS));
  // ultimo elemento corresponde ao coefobj que ja foi processado
  colentr[i] = colentr[i-1];
  // ultima linha da matriz corresponde a uma tautologia
  colentr[i-1] = 1.0;

  FINAL:
  return status;      
}

/* Insere coluna no problema lp, inclui coeficiente da objetiva
// coefobj para nova variavel, o vetor coluna de entrada
// colentr[0..numcnf-1] com o auxilio do vetor com o indice das linhas
// indlin[0..numcnf-1] e do vetor mundo[0..numvar-1] para compor o 
// nome da nova variavel em nomecol do problema lp no ambiente env.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */ 
int atualizalp (CPXCENVptr env, CPXLPptr lp, int numcnf, int numvar, 
                double coefobj, double *colentr, int *indlin,
                int *mundo, char **nomecol) {
  int status = 0;
  int indice = 0; // para compor nome da variavel que entra no lp
  int indcol = 0; // auxiliar na inclusao da nova coluna
  int i;

  /* Obtem indice nos mundos para compor o nome da nova variavel */ 
  for (i = 0; i < numvar; i++)
    indice += (int) (mundo[i]*pow(2,i) + EPS);
  indice++;
  sprintf (nomecol[0], "p%d", indice);
  
  /* Insere nova coluna no lp */
  status = CPXaddcols (env, lp, 1, numcnf, &coefobj, &indcol,
		                   indlin, colentr, NULL, NULL , nomecol);
  if ( status ) {
    printf ("Falhou para inserir coluna.\n");
    goto FINAL;
  }

  FINAL:
  return status;
}

/* Preenche os vetores e variaveis associados ao arquivo de entrada
// cujo nome e nomearq, definindo numero de clausulas numcla_p, numero
// de variaveis logicas numvar_p numero de sentencas numcnf_p, a
// matriz de clausulas CLA_p e o correspondente tamanho de cada
// clausula no vetor tamcla_p, a matriz de sentencas CNF_p e os
// vetores correspondentes ao numero de clausulas na sentenca tamcnf_p
// e o intervalo de probabilidade associado em prob_p e probsup_p.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
// OBS: o arquivo deve conter os dados informados em formato.pdf 
/////////////////////////////////////////////////////////////////// */
int ledados (char *nomearq, int *numcla_p, int *numvar_p,
             int *numcnf_p, int ***CLA_p, int **tamcla_p,
             int ***CNF_p, int **tamcnf_p, double **prob_p,
             double **probsup_p) {
  FILE *arq = NULL, *in = NULL; // para arquivo de entrada e
                                // o sem comentarios
  char barraene;                // para verificar final de linha
  char *arquivo = "limpo.txt";
  int i = 0, j, aux;
  int status = 0;

  arq = fopen (arquivo, "w+"); // escrita e leitura
  in = fopen (nomearq, "r");
  if (arq == NULL || in == NULL) {
    printf ("** ERRO ** Arquivo de entrada nao encontrado.\n");
    status = 1;
    goto FINAL;
  }
  
  /* Elimina comentarios */
  arq = limpa (in, arq);

  /* Leitura dos dados */
  fscanf (arq, "%d %d %d", numvar_p, numcla_p, numcnf_p);
  // alocacao das variaveis relacionadas ao arquivo de entrada
  *CLA_p = (int **) malloc (*numcla_p*sizeof (int *));
  *CNF_p = (int **) malloc (*numcnf_p*sizeof (int *));
  *tamcla_p = (int *) malloc (*numcla_p*sizeof (int));
  *tamcnf_p = (int *) malloc (*numcnf_p*sizeof (int));
  *prob_p = (double *) malloc (*numcnf_p*sizeof (double));
  *probsup_p = (double *) malloc (*numcnf_p*sizeof (double));
  if (*CLA_p == NULL ||*CNF_p == NULL || *tamcla_p == NULL ||
      *tamcnf_p == NULL || *prob_p == NULL || *probsup_p == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  for (i = 0; i < *numcla_p; i++) {
    (*CLA_p)[i] = (int *) malloc (*numvar_p*sizeof (int));
    if ((*CLA_p)[i] == NULL) {
      printf ("** ERRO ** Malloc devolveu NULL!\n");
      status = 1;
      goto FINAL;
    }
  }
  for (i = 0; i < *numcnf_p; i++) {
    (*CNF_p)[i] = (int *) malloc (*numcla_p*sizeof (int));
    if ((*CNF_p)[i] == NULL) {
      printf ("** ERRO ** Malloc devolveu NULL!\n");
      status = 1;
      goto FINAL;
    }
  }
  // leitura e atribuicao de dados
  fscanf (arq, "%c", &barraene);
  if (barraene != '\n') {
    printf ("** ERRO ** Leitura da linha 0 do arquivo de entrada.\n");
    status = 1;
    goto FINAL;
  }
  for (i = 0; i < *numcla_p; i++) {
    for (j = 0, fscanf (arq, "%d", &aux);
         aux != 0; j++, fscanf (arq, "%d", &aux)) {
      (*CLA_p)[i][j] = aux;
      if (abs (aux) > *numvar_p) 
        printf ("** ERRO ** Numero de literal > numvar.\n");
    }
    (*tamcla_p)[i] = j;
    fscanf (arq, "%c", &barraene);
    if (barraene != '\n') {
      printf ("** ERRO ** Leitura da linha %d do arquivo"
              " de entrada.\n", i+1);
      status = 1;
      goto FINAL;
    }
  }
  for (i = 0; i < *numcnf_p; i++) {
    fscanf (arq, "%lf %lf", (*prob_p)+i, (*probsup_p)+i);
    for (j = 0, fscanf (arq, "%d", &aux);
         aux != 0; j++, fscanf (arq, "%d", &aux))
      (*CNF_p)[i][j] = aux;
    (*tamcnf_p)[i] = j; 
    fscanf (arq, "%c", &barraene);
    if (barraene != '\n') {
      printf ("** ERRO ** Leitura da linha %d do arquivo"
              " de entrada.\n", i+1);
      status = 1;
      goto FINAL;
    } 
  }

  FINAL:
  fclose(arq);
  fclose (in);
  remove (arquivo);
  return status;
}

/* Retira comentarios do arquivo in e devolve o endereco para primeira
// posicao do arquivo sem comentarios arq.
/////////////////////////////////////////////////////////////////// */
FILE *limpa (FILE *in, FILE *arq) {
  char aux;
  
  /* Suponho que uma linha de comentarios nao comeca com numero
  // nem sinal de menos (-) */
  do {
    for (aux = getc(in); (aux < '0'|| aux > '9') && aux != '-' 
	     && aux != EOF; aux = getc(in)) 
      while (aux != '\n') aux = getc(in);
    if (aux != EOF) {
      for (aux; aux != '\n'; aux = getc(in))
        putc (aux, arq);
      putc (aux, arq);      
    }
    else break; 
  } while (TRUE);
  rewind (arq);
  return arq;
}

/* Preenche dados iniciais do problema lp (=PMR), construindo 
// um problema inicial fase0 com variaveis de folga
// (z_i = probsup[i]-prob[i], para probsup[i] != prob[i]) e variaveis
// artificiais (y), sendo o lado direito dado pelo vetor
// [probsup[1..nuncnf-2] | 1).
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int preenchelp (CPXENVptr env, CPXLPptr lp, int numcnf, double *prob,
                double *probsup){
  int status = 0;
  FILE *arq = NULL;           // contera problema no formato lp
  char *arquivo = "fase0.lp"; 
  int i, aux, aux1;

  arq = fopen (arquivo, "w");
  if (arq == NULL) {
    printf ("Arquivo %s nao encontrado.", arquivo);
    fclose (arq);
    status = 1;
    goto FINAL;
  }

  /* Controi modelo para fase0  */
  // Funcao objetivo
  fprintf (arq, "\nMinimize\n");
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "+ y%d ", i);
  // Restricoes
  fprintf (arq, "\nSubject to\n");
  aux1 = numcnf-1;
  for (i = 0, aux = 0; i < aux1; i++) {
    if ((probsup[i]- prob[i]) > EPS) {
      fprintf (arq, "y%d + z%d = %f\n", i, aux, probsup[i]);
      aux++;
    }
    else
      fprintf (arq, "y%d = %f\n", i, probsup[i]);
  }
  // Equivale a somatoria das probabilidades dos mundos
  fprintf (arq, "y%d = 1", i);
  /* Limites */
  fprintf (arq, "\nBounds\n");
  for (i = 0, aux = 0; i < aux1; i++)
    if ((probsup[i]- prob[i]) > EPS) {
      fprintf (arq, "0 <= z%d <= %f\n", aux, probsup[i]-prob[i]);
      aux++;
    }
  fprintf (arq, "\nEnd\n");

  fclose (arq);

  /* Le e copia problema para lp  */
  status = CPXreadcopyprob (env, lp, arquivo, NULL);
  if (status)  goto FINAL;

  FINAL:
  remove (arquivo);
  return status;
}

/* Preenche dados iniciais do problema mip (= SP), utilizando as
// matrizes de clausulas CLA[0..numcla-1][] e de sentencas
// CNF[0..numcnf-1][], seus vetores auxiliares de tamanho tamcla e
// tamcnf para aplicarem o processo de linearizacao e definir o
// problema mip com (numcnf+numcla+numvar) variaveis binarias.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int preenchemip (CPXENVptr env, CPXLPptr mip, int numvar, int numcla, 
                 int numcnf, int **CLA, int *tamcla, int **CNF,
                 int *tamcnf) {
  int status = 0;
  FILE *arq = NULL;         // contera problema no formato lp
  int numneg;               // numero de literais negados
  int indcla;               // indice da clausula
  char *arquivo = "mip.lp"; 
  int i, j, k;      

  arq = fopen (arquivo, "w");
  if (arq == NULL) {
    printf ("Arquivo %s nao encontrado.", arquivo);
    fclose (arq);
    status = 1;  
    goto FINAL;
  }

  /* Lineariza as clausulas negadas para gerar problema */
  // Nota: estamos negando cada clausula, logo literal nao
  // negado na clausula aparecera negado ao gerar a restricao.
  // Funcao objetivo
  fprintf (arq, "\nMinimize\n");
  for (i = 0; i < numvar; i++)
    fprintf (arq, "+ 0 x%d ", i);
  // parte relativa a clausulas negadas
  for (i = 0; i < numcla; i++)
    fprintf (arq, "+ 0 y%d ", i);
  // parte relativa as sentencas
  j = numcnf-1;
  for (i = 0; i < j; i++)
    fprintf (arq, "+ s%d ", i);
  // ultima e para sentenca adicional na fase0
  fprintf (arq, "+ 0 s%d ", i);
  // Restricoes
  fprintf (arq, "\nSubject to\n");
  // parte relativa a negacao de clausulas
  for (i = 0; i < numcla; i++) {
    numneg = 0;
    for (k = 0; k < tamcla[i]; k++) {
      if (CLA[i][k] < 0)
	fprintf (arq, "- x%d + y%d <= 0\n", abs (CLA[i][k])-1, i);
      else {
	fprintf (arq, "x%d + y%d <= 1\n", CLA[i][k]-1, i);
	numneg++;
      }
    }
    for (k = 0; k < tamcla[i]; k++) {
      if (CLA[i][k] < 0)
        fprintf (arq, "+ x%d ", abs (CLA[i][k])-1);
      else
        fprintf (arq, "- x%d ", CLA[i][k]-1);
    }
    fprintf (arq, "- y%d <= %d\n", i, (tamcla[i]-(numneg+1)));
  }
  // parte relativa as cnf
  for (i = 0; i < numcnf; i++) {
    for (k = 0; k < tamcnf[i]; k++)
      fprintf (arq, "y%d + s%d <= 1\n", CNF[i][k]-1, i);
    for (k = 0; k < tamcnf[i]; k++)
      fprintf (arq, "- y%d ", CNF[i][k]-1);
    fprintf (arq, "- s%d <= %d\n", i, -1);
  }
  // Tipos
  fprintf (arq, "Binaries\n");
  for (i = 0; i < numvar; i++)
    fprintf (arq, "x%d\n", i);
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "y%d\n", i);
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "s%d\n", i);
  fprintf (arq, "\nEnd\n");

  fclose (arq);

  /* Le e copia problema para mip  */
  status = CPXreadcopyprob (env, mip, arquivo, NULL);
  if (status) {
    printf ("\nErro ao tentar copiar MIP para cplex.\n ");
    goto FINAL;
  }

  FINAL:
  remove (arquivo);
  return status;
}

/* Desaloca ptr e define para NULL
/////////////////////////////////////////////////////////////////// */ 
void desaloca (char **ptr){
  if (*ptr == NULL) {
    free (*ptr);
    *ptr = NULL;
  }
}
