/* Arquivo: geracol.c
// Versao 1.0
//
// Autor: Paulo Sergio de Souza Andrade                n.usp: 4963109
// Fevereiro de 2007
//
// geracol.c 
// Aplicacao para inferencia de um valor min/max de probabilidade para
// uma sentenca adicional, dado como funcao de um banco de dados
// com um conjunto de sentencas na forma normal conjuntiva e seus
// valores de probabilidade dados na forma de intervalos.
// Implementa o metodo de geracao de colunas e utiliza o programa
// de otimizacao GLPK para resolucao dos problemas envolvidos com 
// o metodo.
// Recomenda-se a leitura de formato.pdf para esclarecimento do
// argumento que deve ser passado para o programa, e de Leia-me,
// para instrucoes de compilacao e notas sobre saidas.
/////////////////////////////////////////////////////////////////// */

/* Bibliotecas Utilizadas
/////////////////////////////////////////////////////////////////// */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "glpk.h"

/* Constantes
/////////////////////////////////////////////////////////////////// */
# define TRUE 1
# define MAXLIN 10
# define MAXNOMECOL 1000
# define EPS 1.0e-6
# define ALTTOLOBJ 0.001
# define NORTOLOBJ 1.0e-7

/* Prototipos de Funcoes
/////////////////////////////////////////////////////////////////// */
int preparafase1 (LPX *lp, int numcnf, int *indcoef, int numcoef,
                  double *valorcoef);
int otimizalp (LPX *lp, double *objval_p, double *p, double *pi);
int otimizamip (LPX *mip, int numvar, int numcla, int numcnf,  
                double *x, double *pi, double *objval_p, int *mundo, 
                double *colentr);
int atualizalp (LPX *lp, int numcnf, int numvar, double coefobj,
                double *colentr, int *indlin, int *mundo, 
                char **nomecol);
int leDados (char *nomearq, int *numcla_p, int *numvar_p, 
             int *numcnf_p, int ***CLA_p, int **tamcla_p, 
             int ***CNF_p, int **tamcnf_p, double **prob_p, 
             double **probsup_p);
FILE *limpa (FILE *in, FILE *arq);
int preenchelp (LPX **lp, int numcnf, double *prob, double *probsup);
int preenchemip (LPX **mip, int numvar, int numcla, int numcnf,
                 int **CLA, int *tamcla, int **CNF, int *tamcnf);
void desaloca (char **ptr);

/* Programa Propriamente Dito
/////////////////////////////////////////////////////////////////// */
int main (int numargs, char **v) {  
  int status = 0;
  int i, j;

  /* Dados associados ao arquivo de entrada */
  int numcla;          // numero de clausulas,
  int numvar;          // de variaveis e
  int numcnf;          // de expressoes com probabilidade
  int **CLA = NULL;    // CLA[i][] contera a i-esima clausula
  int *tamcla = NULL;  // tamcla[i] sera o numero de variaveis
                       // que compoem a i-esima clausula
  int **CNF = NULL;    // CNF[i][] contera a i-esima CNF
  int *tamcnf = NULL;  // tamcnf[i] sera o numero de clausulas
                       // que compoem a i-esima CNF
  double *prob = NULL; // prob[i] sera a prob. inferior da i-esima CNF
  double *probsup = NULL; // prob[i] sera a prob. sup. da i-esima CNF
  
  /* Cria ponteiros para problemas do glpk */
  LPX *lp = NULL;   // problema de programacao continua
  LPX *mip = NULL;  // problema de programacao inteira

  /* Vetores e variaveis relacionadas ao problema lp (= PMR) */
  double objval;      // valor otimo do lp
  double *p = NULL;   // vetor solucao de valores do primal
  double *pi = NULL;  // vetor de valores do dual
  int itlp = 0;       // numero de iteracoes para resolver lp
  // para coluna de entrada
  int *indlin = NULL;     // indices das linhas
  double coefobj = 0;     // coeficiente da objetiva
  double *colentr = NULL; // coluna de entrada
  char **nomecol = NULL;  // nome da coluna de entrada
  int *indcoef = NULL;    // atuam na fase0 para guardar indice e
  double *valorcoef = NULL;  // valores da objetiva de novas variveis

  /* Vetores e variaveis relacionadas ao problema mip (= SP) */
  double objvalmip;  // valor otimo
  double *x = NULL;  // vetor solucao
  int itmip = 0;     // numero de iteracoes para resolver mip
  int *mundo = NULL; // atribuicao dada as variaveis {0, 1}
  
  /* Iteraes do mtodo */
  int itgc = 0; // numero de iteracoes do gerador de colunas
  int itex = 0; // numero de vezes que mip foi resolvido exatamente
  
  /* Variaveis para controle do tempo */
  double fim;
  double inicio;
  double tempototallp = 0;
  double tempototalmip = 0;
  time_t t;
  char *horario;

  /* Para arquivo de escrita  */
  FILE *saida = NULL;
  char *arquivo = "saida.txt";

  /* Informacoes iniciais */
  saida = fopen (arquivo, "a");
  if (saida == NULL) {
    printf ("** ERRO ** Arquivo de escrita nao encontrado.\n");
    status = 1;
    goto FINAL;
  }
  t = time(NULL);
  horario = asctime(localtime(&t));
  fprintf (saida,"\nINICIO \nHINI %sARQ %s", horario, v[1]);

  /* Verifica argumentos da linha de comando */
  if (numargs != 2) {
    printf ("** ERRO ** Lista de argumentos errada.\n");
    goto FINAL;
  }
  
  /* Le arquivo de entrada */
  status = leDados (v[1], &numcla, &numvar, &numcnf, &CLA, &tamcla, 
                    &CNF, &tamcnf, &prob, &probsup);
  if (status) goto FINAL;
  // nao preciso da linha abaixo pois tenho uma sent. adicional
  // que foi contabilizada dado que o problema apresenta uma restricao
  // adicional 1*x = 1
  //numcnf++; 
  
  /* Alocacoes */
  // O +1 e devido ao tipo de enderecamento do glpk
  p = (double *) malloc ((numcnf+1)*sizeof (double));
  pi = (double *) malloc ((numcnf+1)*sizeof (double)); 
  indlin = (int *) malloc ((numcnf+1)*sizeof (int));
  colentr = (double *) malloc ((numcnf+1+1)*sizeof (double));
  nomecol = (char **) malloc (sizeof (char *));
  x = (double *) malloc ((numvar+numcla+numcnf+1)*sizeof (double));
  mundo = (int *) malloc (numvar*sizeof (int));
  // Suponho que a fase0 precisara de no maximo numcnf*MAXLIN
  // iteracoes, logo aloco os seguintes vetores
  indcoef = (int *) malloc ((numcnf*MAXLIN)*sizeof (int));
  valorcoef = (double *) malloc ((numcnf*MAXLIN)*sizeof (double));
  if (p == NULL || pi == NULL || indlin == NULL || colentr == NULL ||
      nomecol == NULL || x == NULL || mundo == NULL ||
      indcoef == NULL || valorcoef == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  // Suponho que o nome de uma variavel nao e maior que MAXNOMECOL
  nomecol[0] = (char *) malloc (MAXNOMECOL*sizeof (char));
  if (nomecol[0] == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }

  /* Preenche dados iniciais dos problemas */
  status = preenchelp (&lp, numcnf, prob, probsup) ;
  if (status) {
    printf ("Falhou para criar/preencher LP.\n");
    goto FINAL;
  }

  status = preenchemip (&mip, numvar, numcla, numcnf, CLA, tamcla, 
                        CNF, tamcnf);
  if (status) {
    printf ("Falhou para criar/preencher MIP.\n");
    goto FINAL;
  }

  /* Parte relativa a fase0 do simplex */
  // Altera parametro para resolver mip aproximadamente
  lpx_set_real_parm (mip, LPX_K_TOLOBJ, ALTTOLOBJ);
  
  /* Realiza a fase0 quando i == 0
  // Caso viavel:
  // minimiza para i == 1, fornecendo limite inferior de prob.
  // maximiza para i == 2, fornecendo limite superior de prob.
  ///////////////////////////////////////////////////////////// */
  for (i = 0; i < 3; i++) {
    /* Metodo numerico de geracao de colunas */
    while (TRUE) {
      itgc++;
      /* Otimiza LP e obtem solucao */
      inicio = (double) clock ();
      status = otimizalp (lp, &objval, p, pi);
      fim = (double) clock ();
      tempototallp += fim - inicio;
      if (status) {
        printf ("Falhou para otimizar LP.\n");
        goto FINAL;
      }
      /* Verifica se atingiu 0, implicando problema viavel */
      if (i == 0 && objval < EPS) {
        printf ("Problema e viavel.\n");
        fprintf (saida,"\nRSL S");
        break;
      }
      
      /* Otimiza MIP e obtem solucao */
      inicio = (double) clock ();
      status = otimizamip (mip, numvar, numcla, numcnf, x, pi, 
                           &objvalmip, mundo, colentr);
      fim = (double) clock ();
      tempototalmip += fim - inicio;
      if (status) {
        printf ("Falhou para otimizar MIP.\n");
        goto FINAL;
      }
      
      /* Comprova valor otimo */
      if ((objvalmip > -EPS && (i == 0 || i == 1)) ||
          (objvalmip < EPS && i == 2)) {
        itex++;
        
        /* Altera parametro para definicao padrao */
        lpx_set_real_parm (mip, LPX_K_TOLOBJ, NORTOLOBJ);
        
        /* Realtera multiplicadores */
        for (j = 1; j <= numcnf; j++)
          pi[j] = (-1)*pi[j];
                
        /* Otimiza MIP e obtem solucao */
        inicio = (double) clock ();
        status = otimizamip (mip, numvar, numcla, numcnf, x, pi, 
                             &objvalmip, mundo, colentr);
        fim = (double) clock ();
        tempototalmip += fim - inicio;
        if (status) {
          printf ("Falhou para otimizar MIP.\n");
          goto FINAL;
        }
        
        /* Interpreta resultado */
        // fase0 verificando inviabilidade
        if (i == 0 && objvalmip > -EPS) {
          printf ("Problema e inviavel.\n");
          fprintf (saida,"\nRSL N");
          goto FINAL;
        }
        // finaliza minimizacao
        else if (i == 1 && objvalmip > -EPS) {
          printf ("limite inferior = %g\n", objval); 
          fprintf (saida,"\nMIN %e", objval);
          break;
        }
        // finaliza maximizacao
        else if (objvalmip < EPS){
          printf ("limite superior = %g\n", objval); 
          fprintf (saida,"\nMAX %e", objval);
          break;  
        } 

        /* Realtera parametro para resolver mip aproximadamente */
        lpx_set_real_parm (mip, LPX_K_TOLOBJ, ALTTOLOBJ);
      }
      
      /* Coleta indice do coeficiente da objetiva */
      // na fase0 guardo para que na fase1 voltem
      if (i == 0) {
       indcoef[itgc] = lpx_get_num_cols (lp);
       indcoef[itgc]++;
       valorcoef[itgc] = colentr[numcnf+1];
      } 
      // coefobj so foi zero na fase0
      else coefobj = colentr[numcnf+1]; 

      /*  Insere coluna no lp */
      status = atualizalp (lp, numcnf, numvar, coefobj, colentr, 
                                indlin, mundo, nomecol);
      if (status) {
        printf ("Falhou para atualizar simplex.\n");
        goto FINAL;
      }
    }
    
    /* Termino da fase0 com sucesso => problema viavel, aplico a fase
    // seguinte, minimizacao seguida de maximixacao */
    if (i == 0) {
      status = preparafase1 (lp, numcnf, indcoef, itgc-1, valorcoef); 
      if (status) {
        printf ("Erro ao preparar problema para fase1.\n");
        goto FINAL;
      }
      /* Altera coeficientes da objetiva para mip considerar
      // sentenca adicional */
      lpx_set_obj_coef (mip, numvar+numcla+numcnf, 1.0);   
    }
    else if (i == 1) {
      lpx_set_obj_dir (lp, LPX_MAX);
      lpx_set_obj_dir (mip, LPX_MAX);     
    }
  }

  FINAL:
    
  ////////////////////////////////////////////////////////////
  /* Curiosidade
  if (0 != lpx_print_sol(lp, "Solucaolp.txt"))
    printf ("Falhou para obter arquivo com solucao.\n");
  if (0 != lpx_write_bas(lp, "base.txt"))
    printf ("Falhou para obter arquivo com base.\n");
  if (0 != lpx_write_cpxlp (lp, "PSAT.txt"))
    printf ("Falhou para obter arquivo com problema LP.\n");
  ////////////////////////////////////////////////////////// */  
  
  /* Interrupcao da fase0 por algum motivo, aconselhavel
  // verificar informacao enviada para tela */
  if (status)
    fprintf (saida,"\nRSL F");
      
  /* Envia informacoes finais ao arquivo de saida */
  fprintf (saida, "\nTMPC %e", tempototallp/CLOCKS_PER_SEC);
  fprintf (saida, "\nTMPI %e", tempototalmip/CLOCKS_PER_SEC);
  // note que nao coletei informacoes sobre iteracoes do mip e do lp
  fprintf (saida, "\nITGC %d", itgc);
  fprintf (saida, "\nITEX %d", itex);

  /* Desaloca problemas */
  if (lp != NULL) lpx_delete_prob (lp);
 
  if (mip != NULL) lpx_delete_prob (mip);

  /* Desaloca vetores e matrizes de dados */
  desaloca ((char **) &p);
  desaloca ((char **) &pi);
  desaloca ((char **) &indlin);
  desaloca ((char **) &colentr);
  desaloca ((char **) &nomecol[0]);
  desaloca ((char **) &nomecol);
  desaloca ((char **) &x);
  desaloca ((char **) &mundo);
  desaloca ((char **) &indcoef);
  desaloca ((char **) &valorcoef);
  if (CLA != NULL) {
    for (i = 0; i < numcla; i++)
      desaloca ((char **) &CLA[i]);
  }
  desaloca ((char **) &CLA);
  if (CNF != NULL) {
    for (i = 0; i < numcnf; i++)
      desaloca ((char **) &CNF[i]);
  }
  desaloca ((char **) &CNF);
  desaloca ((char **) &tamcla);
  desaloca ((char **) &tamcnf);
  desaloca ((char **) &prob);
  desaloca ((char **) &probsup);
  
  /* Informacoes finais */
  t = time(NULL);
  horario = asctime(localtime(&t));
  fprintf (saida,"\nHFIM %sFIM", horario);

  fclose (saida);

  return EXIT_SUCCESS;
}


/* Altera funcao objetivo com os coeficientes de colunas geradas, que 
// haviam ficado de fora durante a fase0, sendo que o valor dos mesmos
// esta em valorcoef[1...numcoef] e seus indices em indcoef, em
// seguida elimina de lp, variavel nao basica, diferente de variavel
// de folga.
// Devolve 1 se uma falha ocorre e 0 caso contrario. 
/////////////////////////////////////////////////////////////////// */
int preparafase1 (LPX *lp, int numcnf, int *indcoef, int numcoef,
                    double *valorcoef) {
  int status = 0;
  int numcol; // numero de colunas no problema
  int cstat;  // status da coluna, basica ou nao basica
  int *inddelcol = NULL; // indices de variaveis nao basicas,
                         // diferentes de variaveis de folga
  int numdel = 0; // numero de colunas com a caracteristica acima que
                  // serao eliminadas do problema
  char *nomecol;  // ponteiro para nome de variavel
  int i; 

  numcol = lpx_get_num_cols (lp);
  inddelcol = (int *) malloc (numcol*sizeof (int));
  if (inddelcol == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  // identifica variavel nao basica e diferente de var. de folga
  for (i = 1; i <= numcol; i++)
    if (lpx_get_col_stat (lp, i) != LPX_BS) {
      nomecol = lpx_get_col_name (lp, i);
      if (nomecol[0] != 'z') {
        numdel++;
        inddelcol[numdel] = i; 
      }  
    }
    
  // reconstroi funcao objetiva para fase1  
  for (i = 1; i <= numcoef; i++)
    lpx_set_obj_coef (lp, indcoef[i], valorcoef[i]);
    
  // elimina colunas
  lpx_del_cols(lp, numdel, inddelcol);
  
  // redefine limites e coeficiente da funcao objetivo sobre variavel
  // artificial caso a mesma esteja na base
  numcol = lpx_get_num_cols (lp);
  for (i = 1; i <= numcol; i++) {
    nomecol = lpx_get_col_name (lp, i);
    if (nomecol[0] == 'y') {
      lpx_set_obj_coef (lp, i, 0.0);
      lpx_set_col_bnds(lp, i, LPX_FX, 0.0, 0.0);
    }
  }
  
  FINAL:
  desaloca ((char **) &inddelcol);
  return status;                            
}

/* Otimiza o problema lp, fornecendo o valor otimo da objetiva em
// objval_P, o vetor p com os valores das variaveis que compoem a
// base, e o vetor pi de valores do dual.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */ 
int otimizalp (LPX *lp, double *objval_p, double *p, double *pi) {
  int status = 0;
  int numlin, i; // numero de linhas

  /* Otimiza LP e obtem solucao */
  if (lpx_simplex(lp) != LPX_E_OK) {
    printf ("Falhou para otimizar LP.\n");
    status = 1;
    goto FINAL;
  }
  /* OBS: Obter numero de colunas e alocar p caso
  // haja interesse em obte-lo */
  if (lpx_get_status(lp) != LPX_OPT && 
      lpx_get_status(lp) != LPX_FEAS) {
    printf ("Solucao do LP nao e otima nem viavel.\n");
    status = 1;
    goto FINAL;
  }
  *objval_p = lpx_get_obj_val(lp);
  numlin = lpx_get_num_rows(lp);
  for (i = 1; i <= numlin; i++)
    pi[i] = lpx_get_row_dual(lp, i);

  FINAL:
  return status;
}

/* Otimiza subproblema mip, porem antes altera coeficientes da funcao
// objetivo com os vetores pi[1..nuncnf] e indices em [(numvar+numcla
// +1)..(numvar+numcla+numcnf)], e fornece o valor otimo em objval_p,
// a solucao otima titulada no vetor x[1..(numvar+numcnf)], a que 
// define o mundo possivel no vetor  mundo[1..numvar] e a coluna de
// entrada em colentr[1..numcnf].
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int otimizamip (LPX *mip, int numvar, int numcla, int numcnf,  
                double *x, double *pi, double *objval_p, int *mundo, 
                double *colentr) {
  int status = 0;
  int solstat;  // status da solucao
  int numcol;   // numero de colunas no problema
  double cte = 0; // constante associada ao mip
  int i, aux;

  /* Prepara multiplicadores */
  for (i = 1; i <= numcnf; i++)
    pi[i] = (-1)*pi[i];
  // constante
  cte = pi[numcnf];
  /* Altera coeficientes da funcao objetivo */
  // primeiro indice dos que precisam ser alterados
  aux = numvar + numcla;
  for (i = 1; i < numcnf; i++) 
    lpx_set_obj_coef (mip, aux+i, pi[i]);
  /* Otimiza o problema e obtem solucao */
  if (lpx_intopt (mip) != LPX_E_OK) {
    printf ("Falhou para otimizar MIP.\n");
    status = 1;
    goto FINAL;
  }
  if (lpx_mip_status (mip) != LPX_I_OPT && 
      lpx_mip_status (mip) != LPX_I_FEAS) {
    printf ("MIP nao forneceu solucao.\n");
    status = 1;
    goto FINAL;
  }
  *objval_p = lpx_mip_obj_val (mip);
  /* soma constante que havia ficado de fora */
  *objval_p += cte;
  numcol = lpx_get_num_cols (mip);
  for (i = 1; i <= numcol; i++)
    x[i] = lpx_mip_col_val (mip, i);

  // obtem mundo
  for (i = 0; i < numvar; i++)
    mundo[i] = (int) (x[i+1] + EPS);
    
  /* Obtem coluna de entrada resultante */
  for (i = 1; i <= numcnf; i++)
    colentr[i] = (double)((int) (x[i+aux] + EPS));
  // ultimo elemento corresponde ao coefobj que ja foi processado
  colentr[i] = colentr[i-1];
  // ultima linha da matriz corresponde a uma tautologia
  colentr[i-1] = 1.0;

  FINAL:
  return status;      
}
/* Insere coluna no problema lp, inclui coeficiente da funcao objetivo
// coefobj para nova variavel, o vetor coluna de entrada
// colentr[1..numcnf] com o auxilio do vetor com o indice das linhas
// indlin[1..numcnf] e do vetor mundo[1..numvar] para compor o 
// nome da nova variavel em nomecol.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */ 
int atualizalp (LPX *lp, int numcnf, int numvar, double coefobj, 
                double *colentr, int *indlin, int *mundo, 
                char **nomecol) {
  int status = 0;
  int indice = 0; // para compor nome da variavel que entra no lp
  int numcol = 1; // numero de novas colunas
  int numnzs = 0; // numero de coeficientes != de 0 na nova coluna
  int indcol;     // auxiliar na inclusao da nova coluna
  int i;

  /* Obtem indice nos mundos para compor o nome da nova variavel */
  for (i = 0; i < numvar; i++)
    indice += (int) (mundo[i]*pow(2,i) + EPS);
  indice++;
  sprintf (nomecol[0], "p%d", indice);
  // adiciona coluna e devolve indice
  indcol = lpx_add_cols(lp, numcol);
  // indices das linhas com coef != 0 para lp adicionar coluna
  for (i = 1; i <= numcnf; i++) {
    if (colentr[i] > EPS) {
      indlin[numnzs+1] = i;
      colentr[numnzs+1] = 1.0;
      numnzs++;
    }
  }
  
  /* Altera nome, coeficientes e limites da coluna de entrada */ 
  lpx_set_mat_col (lp, indcol , numnzs, indlin, colentr);
  lpx_set_obj_coef (lp, indcol, coefobj);
  lpx_set_col_name (lp, indcol, nomecol[0]);  
  lpx_set_col_bnds(lp, indcol, LPX_LO, 0.0, 0.0);
  
  FINAL:
  return status;
}

/* Preenche os vetores e variaveis associados ao arquivo de entrada
// cujo nome e nomearq, definindo numero de clausulas numcla_p, numero
// de variaveis logicas numvar_p numero de sentencas numcnf_p, a
// matriz de clausulas CLA_p e o correspondente tamanho de cada
// clausula no vetor tamcla_p, a matriz de sentencas CNF_p e os
// vetores correspondentes ao numero de clausulas na sentenca tamcnf_p
// e o intervalo de probabilidade associado em prob_p e probsup_p.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
// OBS: o arquivo deve conter os dados informados em formato.pdf. 
/////////////////////////////////////////////////////////////////// */
int leDados (char *nomearq, int *numcla_p, int *numvar_p,
             int *numcnf_p, int ***CLA_p, int **tamcla_p,
             int ***CNF_p, int **tamcnf_p, double **prob_p,
             double **probsup_p) {
  FILE *arq = NULL, *in = NULL; // para arquivo de entrada e
                                // o sem comentarios
  char barraene;                // para verificar final de linha
  char *arquivo = "limpo.txt";
  int i = 0, j, aux;
  int status = 0;

  arq = fopen (arquivo, "w+"); // escrita e leitura
  in = fopen (nomearq, "r");
  if (arq == NULL || in == NULL) {
    printf ("** ERRO ** Arquivo de entrada nao encontrado.\n");
    status = 1;
    goto FINAL;
  }
  
  /* Elimina comentarios */
  arq = limpa (in, arq);

  /* Leitura dos dados */
  fscanf (arq, "%d %d %d", numvar_p, numcla_p, numcnf_p);
  // alocacao das variaveis relacionadas ao arquivo de entrada
  *CLA_p = (int **) malloc (*numcla_p*sizeof (int *));
  *CNF_p = (int **) malloc (*numcnf_p*sizeof (int *));
  *tamcla_p = (int *) malloc (*numcla_p*sizeof (int));
  *tamcnf_p = (int *) malloc (*numcnf_p*sizeof (int));
  *prob_p = (double *) malloc (*numcnf_p*sizeof (double));
  *probsup_p = (double *) malloc (*numcnf_p*sizeof (double));
  if (*CLA_p == NULL ||*CNF_p == NULL || *tamcla_p == NULL ||
      *tamcnf_p == NULL || *prob_p == NULL || *probsup_p == NULL) {
    printf ("** ERRO ** Malloc devolveu NULL!\n");
    status = 1;
    goto FINAL;
  }
  for (i = 0; i < *numcla_p; i++) {
    (*CLA_p)[i] = (int *) malloc (*numvar_p*sizeof (int));
    if ((*CLA_p)[i] == NULL) {
      printf ("** ERRO ** Malloc devolveu NULL!\n");
      status = 1;
      goto FINAL;
    }
  }
  for (i = 0; i < *numcnf_p; i++) {
    (*CNF_p)[i] = (int *) malloc (*numcla_p*sizeof (int));
    if ((*CNF_p)[i] == NULL) {
      printf ("** ERRO ** Malloc devolveu NULL!\n");
      status = 1;
      goto FINAL;
    }
  }
  // leitura e atribuicao de dados
  fscanf (arq, "%c", &barraene);
  if (barraene != '\n') {
    printf ("** ERRO ** Leitura da linha 0 do arquivo de entrada.\n");
    status = 1;
    goto FINAL;
  }
  for (i = 0; i < *numcla_p; i++) {
    for (j = 0, fscanf (arq, "%d", &aux);
         aux != 0; j++, fscanf (arq, "%d", &aux)) {
      (*CLA_p)[i][j] = aux;
      if (abs (aux) > *numvar_p) 
        printf ("** ERRO ** Numero de literal > numvar.\n");
    }
    (*tamcla_p)[i] = j;
    fscanf (arq, "%c", &barraene);
    if (barraene != '\n') {
      printf ("** ERRO ** Leitura da linha %d do arquivo"
              " de entrada.\n", i+1);
      status = 1;
      goto FINAL;
    }
  }
  for (i = 0; i < *numcnf_p; i++) {
    fscanf (arq, "%lf %lf", (*prob_p)+i, (*probsup_p)+i);
    for (j = 0, fscanf (arq, "%d", &aux);
         aux != 0; j++, fscanf (arq, "%d", &aux))
      (*CNF_p)[i][j] = aux;
    (*tamcnf_p)[i] = j; 
    fscanf (arq, "%c", &barraene);
    if (barraene != '\n') {
      printf ("** ERRO ** Leitura da linha %d do arquivo"
              " de entrada.\n", i+1);
      status = 1;
      goto FINAL;
    } 
  }

  FINAL:
  fclose(arq);
  fclose (in);
  remove (arquivo);
  return status;
}

/* Retira comentarios do arquivo in e devolve o endereco para primeira
// posicao do arquivo sem comentarios arq.
/////////////////////////////////////////////////////////////////// */
FILE *limpa (FILE *in, FILE *arq) {
  char aux;
  
  /* Suponho que uma linha de comentarios nao comeca com numero
  // nem sinal de menos (-) */
  do {
    for (aux = getc(in); (aux < '0'|| aux > '9') && aux != '-' 
	   && aux != EOF; aux = getc(in)) 
      while (aux != '\n') aux = getc(in);
    if (aux != EOF) {
      for (aux; aux != '\n'; aux = getc(in))
        putc (aux, arq);
      putc (aux, arq);      
    }
    else break; 
  } while (TRUE);
  rewind (arq);
  return arq;
}

/* Preenche dados iniciais do problema lp (=PMR), construindo 
// um problema inicial fase0 com variaveis de folga 
// (z_i = probsup[i]-prob[i], para probsup[i] != prob[i]) e variaveis
// artificiais (y), sendo o lado direito dado pelo vetor 
// [probsup[1..nuncnf-2] | 1). Todas variaveis sao do tipo >= 0.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int preenchelp (LPX **lp, int numcnf, double *prob, double *probsup) {
  int status = 0;
  FILE *arq = NULL;  // contera problema no formato lp
  char *arquivo = "fase0.lp";
  int i, aux, aux1;

  arq = fopen (arquivo, "w");
  if (arq == NULL) {
    printf ("Arquivo %s nao encontrado.", arquivo);
    fclose (arq);
    status = 1;
    goto FINAL;
  }

  /* Controi modelo para fase0  */
  // Funcao objetivo
  fprintf (arq, "\nMinimize\n");
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "+ y%d ", i);
  // Restricoes
  fprintf (arq, "\nSubject to\n");
  aux1 = numcnf-1;
  for (i = 0, aux = 0; i < aux1; i++) {
    if ((probsup[i]- prob[i]) > EPS) {
      fprintf (arq, "y%d + z%d = %f\n", i, aux, probsup[i]);
      aux++;
    }
    else 
      fprintf (arq, "y%d = %f\n", i, probsup[i]);
  }
  // Equivale a somatoria das probabilidades dos mundos  
  fprintf (arq, "y%d = 1", i);
  /* Limites */
  fprintf (arq, "\nBounds\n");
  for (i = 0, aux = 0; i < aux1; i++)
    if ((probsup[i]- prob[i]) > EPS) {  
      fprintf (arq, "0 <= z%d <= %f\n", aux, probsup[i]-prob[i]);
      aux++;
    }
  fprintf (arq, "\nEnd\n");

  fclose (arq);

  /* Le e copia problema para lp  */
  *lp = lpx_read_cpxlp (arquivo);
  if (*lp == NULL) status = 1;
  if ( status ) goto FINAL;
  lpx_set_prob_name(*lp, "PSAT");
 
  FINAL:
  remove (arquivo);
  return status;
}

/* Preenche dados iniciais do problema mip (= SP), utilizando as
// matrizes de clausulas CLA[0..numcla-1][] e de sentencas 
// CNF[0..numcnf-1][], seus vetores auxiliares de tamanho tamcla e 
// tamcnf para aplicarem o processo de linearizacao e definir o 
// problema mip com (numcnf+numcla+numvar) variaveis binarias.
// Devolve 1 se uma falha ocorre e 0 caso contrario.
/////////////////////////////////////////////////////////////////// */
int preenchemip (LPX **mip, int numvar, int numcla, int numcnf,
                 int **CLA, int *tamcla, int **CNF, int *tamcnf) {
  int status = 0;
  FILE *arq = NULL; // contera problema no formato lp
  int numneg;       // numero de literais negados
  int indcla;       // indice da clausula
  char *arquivo = "mip.lp"; 
  int i, j, k;      

  arq = fopen (arquivo, "w");
  if (arq == NULL) {
    printf ("Arquivo %s nao encontrado.", arquivo);
    fclose (arq);
    status = 1;  
    goto FINAL;
  }

  /* Lineariza as clausulas negadas para gerar problema */
  // Nota: estamos negando cada clausula, logo literal nao
  // negado na clausula aparecera negado ao gerar a restricao.
  // Funcao objetivo
  fprintf (arq, "\nMinimize\n");
  for (i = 0; i < numvar; i++)
    fprintf (arq, "+ 0 x%d ", i);
  // parte relativa a clausulas negadas
  for (i = 0; i < numcla; i++)
    fprintf (arq, "+ 0 y%d ", i);
  // parte relativa as sentencas
  j = numcnf-1; 
  for (i = 0; i < j; i++)
    fprintf (arq, "+ s%d ", i);
  // ultima e para sentenca adicional na fase0
  fprintf (arq, "+ 0 s%d ", i);
  // Restricoes
  fprintf (arq, "\nSubject to\n");
  // parte relativa a negacao de clausulas
  for (i = 0; i < numcla; i++) {
    numneg = 0;
    for (k = 0; k < tamcla[i]; k++) {
	  if (CLA[i][k] < 0)
	    fprintf (arq, "- x%d + y%d <= 0\n", abs (CLA[i][k])-1, i);
      else {
	    fprintf (arq, "x%d + y%d <= 1\n", CLA[i][k]-1, i);
	    numneg++; 
      }
    }
    for (k = 0; k < tamcla[i]; k++) {
	  if (CLA[i][k] < 0)
        fprintf (arq, "+ x%d ", abs (CLA[i][k])-1);
      else
        fprintf (arq, "- x%d ", CLA[i][k]-1);
    }
    fprintf (arq, "- y%d <= %d\n", i, (tamcla[i]-(numneg+1)));
  }
  // parte relativa as cnf
  for (i = 0; i < numcnf; i++) {
    for (k = 0; k < tamcnf[i]; k++)
	  fprintf (arq, "y%d + s%d <= 1\n", CNF[i][k]-1, i);
    for (k = 0; k < tamcnf[i]; k++)
        fprintf (arq, "- y%d ", CNF[i][k]-1);
    fprintf (arq, "- s%d <= %d\n", i, -1);
  }
  // Tipos
  fprintf (arq, "Binaries\n");
  for (i = 0; i < numvar; i++)
    fprintf (arq, "x%d\n", i);
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "y%d\n", i);
  for (i = 0; i < numcnf; i++)
    fprintf (arq, "s%d\n", i);
  fprintf (arq, "\nEnd\n");

  fclose (arq);

  /* Le e copia problema para mip  */
  *mip = lpx_read_cpxlp (arquivo);
  if (*mip == NULL) {
    status = 1;
    goto FINAL;
  }
  lpx_set_prob_name(*mip, "Subprob");
  
  FINAL:
  remove (arquivo);
  return status;
}

/* Desaloca *ptr e define para NULL
/////////////////////////////////////////////////////////////////// */ 
void desaloca (char **ptr){
  if (*ptr == NULL) {
    free (*ptr);
    *ptr = NULL;
  }
}
